/* game/queue.c */
/* command queue processing */

#include "externs.h"

int queue_size[NUM_QUEUES];

static struct bque *qfirst, *qlast;
static struct bque *qwait, *qwaitlast;

long next_wait;


void immediate_que(dbref player, char *cmd, dbref thing)
{
  char buf[8192], buff[strlen(cmd)+1], *r, *s=buff;
  char *sptr[10];
  int a, len[10];

  /* check that the player can do a queue command */
  if((db[player].flags & (GOING|HAVEN)) || Typeof(player) == TYPE_GARBAGE)
    return;

  /* Preserve environment %0 to %9 */
  for(a=0;a<10;a++) {
    len[a]=strlen(wptr[a])+1;
    memcpy(sptr[a]=alloca(len[a]), wptr[a], len[a]);
  }

  strcpy(buff, cmd);
  while((r=ansi_parse_up(&s, ';'))) {
    func_zerolev();
    pronoun_substitute(buf, player, thing, r);
    if(*buf)
      process_command(player, buf, thing, 0);

    /* Restore local registers after each command */
    for(a=0;a<10;a++)
      memcpy(wptr[a], sptr[a], len[a]);

    /* Exit if object destroyed itself */
    if(Typeof(player) == TYPE_GARBAGE)
      return;
  }
}

void parse_que(dbref player, char *text, dbref cause)
{
  struct bque *new;
  char buff[strlen(text)+1], *cmd, *s=buff;
  int a=0, len, pow_queue=power(player, POW_QUEUE);

  if((db[player].flags & (GOING|HAVEN)) || Typeof(player) == TYPE_GARBAGE)
    return;

  strcpy(buff, text);

  /* add commands to queue */
  while((cmd=ansi_parse_up(&s, ';'))) {
    /* make sure player can afford to do it */
    if(!pow_queue) {
      if(db[player].queue >= QUEUE_QUOTA) {
        db[player].flags |= HAVEN;
        notify(db[player].owner, "Run away object: %s",
               unparse_object(db[player].owner, player));
        return;
      } if(!db[db[player].owner].pennies &&
           !charge_money(player, rand_num(QUEUE_COST)?0:1, 0)) {
        notify(db[player].owner, "Not enough money to queue command on %s.",
               unparse_object(db[player].owner, player));
        return;
      }
    }

    if(queue_size[Q_COMMAND] >= (pow_queue?MAX_WIZARD_CMDS:MAX_PLAYER_CMDS)) {
      notify(db[player].owner, "%s: Too many queued commands in progress.",
             unparse_object(db[player].owner, player));
      return;
    }

    /* create pointer reference */
    len=strlen(cmd)+1;
    new=malloc(sizeof(struct bque)+len);
    new->next=NULL;
    new->player=player;
    new->cause=cause;
    for(a=0;a<10;a++) {
      if(*wptr[a])
        SPTR(new->env[a], wptr[a]);
      else
        new->env[a]=NULL;
    }
    memcpy(new->cmd, cmd, len);

    /* add to the queue list */
    if(qlast) {
      qlast->next=new;
      qlast=new; 
    } else
      qlast=qfirst=new;

    db[player].queue++;
    queue_size[Q_COMMAND]++;
  }
}

void do_wait(player, arg1, command, argv, pure, cause)
dbref player, cause;
char *arg1, *command, **argv, *pure;
{
  struct bque *new;
  int a, len, wait=atoi(arg1), pow_queue=power(player, POW_QUEUE);

  if(db[player].flags & (GOING|HAVEN))
    return;

  /* execute command immediately if no time is specified */
  if(wait < 1) {
    parse_que(player, command, cause);
    return;
  } else if(wait > 31536000)  /* 1 year maximum */
    wait=31536000;

  /* make sure player can afford to do it */
  if(!pow_queue) {
    if(db[player].queue >= QUEUE_QUOTA) {
      db[player].flags |= HAVEN;
      notify(db[player].owner, "Run away object: %s",
             unparse_object(db[player].owner, player));
      return;
    } if(!db[db[player].owner].pennies) {
      notify(db[player].owner, "Not enough money to queue @wait on %s.",
             unparse_object(db[player].owner, player));
      return;
    }
  }

  if(queue_size[Q_WAIT] >= (pow_queue?MAX_WIZARD_WAIT:MAX_PLAYER_WAIT)) {
    notify(db[player].owner, "%s: Too many @wait commands in progress.",
           unparse_object(db[player].owner, player));
    return;
  }

  /* create pointer reference */
  len=strlen(command)+1;
  new=malloc(sizeof(struct bque)+len);
  new->next=NULL;
  new->player=player;
  new->cause=cause;
  new->wait=now+wait;
  for(a=0;a<10;a++) {
    if(*wptr[a])
      SPTR(new->env[a], wptr[a]);
    else
      new->env[a]=NULL;
  }
  memcpy(new->cmd, command, len);

  /* add to the @wait queue */
  if(qwaitlast) {
    qwaitlast->next=new;
    qwaitlast=new; 
  } else
    qwaitlast=qwait=new;

  /* new value for @wait queue to wake up on */
  if(!next_wait || new->wait < next_wait)
    next_wait=new->wait;

  db[player].queue++;
  queue_size[Q_WAIT]++;
}

/* execute one round of queue commands */
void flush_queue()
{
  struct bque *ptr;
  int a, i, num=queue_size[Q_COMMAND];
  char buff[8192];

  if(!num)
    return;

  nstat[NS_QUEUE]++;
  for(i=0;i<num && qfirst;i++) {
    /* check time slice */
    if(game_online && get_time() > time_slice) {
      nstat[NS_OVERRUNS]++;
      break;
    }

    /* remove entry from queue */
    ptr=qfirst;
    if(!(qfirst=ptr->next))
      qlast=NULL;
    queue_size[Q_COMMAND]--;

    /* process command entry */
    db[ptr->player].queue--;
    if(!(db[ptr->player].flags & (GOING|HAVEN))) {
      for(a=0;a<10;a++)
        strcpy(wptr[a], ptr->env[a]?:"");
      func_zerolev();
      pronoun_substitute(buff, ptr->player, ptr->cause, ptr->cmd);
      if(*buff)
        process_command(ptr->player, buff, ptr->cause, 0);
    }

    /* remove pointer */
    for(a=0;a<10;a++)
      free(ptr->env[a]);
    free(ptr);
  }

  for(a=0;a<10;a++)
    *wptr[a]='\0';
}

/* called every second to check for wait queue commands */           
void wait_second()
{
  struct bque *ptr, *next, *last=NULL;
  int a;

  if(!next_wait || next_wait > now)  /* nothing to do yet */
    return;
  next_wait=0;

  for(ptr=qwait;ptr;ptr=next) {
    next=ptr->next;
    if(ptr->wait > now && !(db[ptr->player].flags & (GOING|HAVEN))) {
      last=ptr;
      if(!next_wait || ptr->wait < next_wait)
        next_wait=ptr->wait;
      continue;
    }

    /* remove from queue */
    if(last)
      last->next=ptr->next;
    else
      qwait=ptr->next;
    queue_size[Q_WAIT]--;

    /* process @wait command */
    db[ptr->player].queue--;
    if(!(db[ptr->player].flags & (GOING|HAVEN))) {
      for(a=0;a<10;a++)
        strcpy(wptr[a], ptr->env[a]?:"");
      parse_que(ptr->player, ptr->cmd, ptr->cause);
    }

    for(a=0;a<10;a++)
      free(ptr->env[a]);
    free(ptr);
  }
  qwaitlast=last;
}

/* tell player what commands they have pending in the queue */
void do_queue(dbref player, char *arg1, char *arg2)
{
  const char *options[]={"All", "Immediate", "Waited", 0};

  struct bque *tmp;
  dbref thing=power(player, POW_QUEUE)?AMBIGUOUS:player;
  int type=0, count, tcount=0;
  
  /* Select a type of viewing */
  if(*arg1) {
    for(type=0;options[type];type++)
      if(string_prefix(options[type], arg1))
        break;
    if(!options[type]) {
      notify(player, "No such queue '%s'.", arg1);
      return;
    }
  }

  /* If present, only view the queue on <arg2> */
  if(*arg2) {
    if((thing=match_thing(player, arg2, MATCH_EVERYTHING)) == NOTHING)
      return;
    if(!controls(player, thing, POW_QUEUE)) {
      notify(player, "Permission denied.");
      return;
    }
  }

  /* Immediate commands */
  if(type < 2) {
    for(count=0,tmp=qfirst;tmp;tmp=tmp->next)
      if((thing == AMBIGUOUS || thing == tmp->player ||
          thing == db[tmp->player].owner) &&
         controls(player, tmp->player, POW_QUEUE)) {
        if(!count)
          notify(player, "\e[1mImmediate commands:\e[m");
        notify(player, "%s:%s", unparse_object(player, tmp->player), tmp->cmd);
        count++, tcount++;
      }
    if(count > 0)
      notify(player, "-- Total: %d Immediate --", count);
  }

  /* Waited Commands */
  if(!type || type == 2) {
    for(count=0,tmp=qwait;tmp;tmp=tmp->next)
      if((thing == AMBIGUOUS ||
          thing == tmp->player || thing == db[tmp->player].owner) &&
         controls(player, tmp->player, POW_QUEUE)) {
        if(!count)
          notify(player, "\e[1m@waited commands:\e[m");
        notify(player, "\e[1;36m[\e[37m%ld\e[36m]\e[m%s:%s", 
               tmp->wait-now, unparse_object(player, tmp->player), tmp->cmd);
        count++, tcount++;
      }
    if(count > 0)
      notify(player, "-- Total: %d @waited --", count);
  }

  if(!tcount)
    notify(player, "Nothing in the queue.");
}

/* remove all queued commands from a certain player */
void do_halt(player, arg1, ncom, argv, pure, cause)
dbref player, cause;
char *arg1, *ncom, **argv, *pure;
{
  int thing=player;

  if(!strcmp(arg1, "ALL") && power(player, POW_SECURITY))
    thing=AMBIGUOUS;
  else if(*arg1) {
    if((thing=match_thing(player, arg1, MATCH_EVERYTHING)) == NOTHING)
      return;
    if(!controls(player, thing, POW_MODIFY)) {
      notify(player, "Permission denied.");
      return;
    }
  }

  if(!Quiet(player)) {
    if(thing == player)
      notify(player, "Everything halted.");
    else if(thing == AMBIGUOUS)
      notify(player, "Halted away every blessed thing.");
    else
      notify(player, "%s halted.", unparse_object(player, thing));
  }

  /* remove the queue commands */
  halt_object(thing);

  /* execute extra command? */
  if(*ncom && !Invalid(thing))
    parse_que(thing,ncom,cause);
}

void halt_object(dbref thing)
{
  struct bque *ptr, *next, *last;
  int a, queue;

  for(queue=0;queue < 2;queue++) {
    last=NULL;
    for(ptr=(queue == Q_COMMAND)?qfirst:qwait; ptr; ptr=next) {
      next=ptr->next;
      if(thing != AMBIGUOUS && ptr->player != thing &&
         db[ptr->player].owner != thing) {
        last=ptr;
        continue;
      }

      /* remove entry from queue */
      if(last)
        last->next=ptr->next;
      else
        (queue == Q_COMMAND)?qfirst:qwait = ptr->next;

      queue_size[queue]--;
      db[ptr->player].queue--;

      /* free command entry */
      for(a=0;a<10;a++)
        free(ptr->env[a]);
      free(ptr);
    }

    /* set the tail pointer */
    (queue == Q_COMMAND)?qlast:qwaitlast = last;
  }
}


void save_queues(FILE *f)
{
  struct bque *ptr;
  int i, j;

  /* Save Command Queues */
  for(j=0;j<2;j++) {
    if((ptr=(j == Q_COMMAND)?qfirst:qwait)) {
      putchr(f, j+1);

      for(;ptr;ptr=ptr->next) {
        putnum(f, ptr->player);
        putstring(f, ptr->cmd);
        putnum(f, ptr->cause);
        if(j == Q_WAIT)
          putlong(f, ptr->wait);
        for(i=0;i<10;i++)
          putstring(f, ptr->env[i]?:"");
      }
      putnum(f, NOTHING);
    }
  }

  putchr(f, 0);
}

void load_queues(FILE *f)
{
  struct bque *new, *last;
  int i, len, count, queue;
  dbref obj;
  char *s;

  while((queue=fgetc(f)) && !feof(f)) {
    count=0;
    switch(--queue) {
    case Q_COMMAND:
    case Q_WAIT:
      last=NULL;
      while((obj=getnum(f)) != NOTHING && !feof(f)) {
        s=getstring_noalloc(f);
        len=strlen(s)+1;
        new=malloc(sizeof(struct bque)+len);
        new->next=NULL;
        new->player=obj;
        memcpy(new->cmd, s, len);
        new->cause=getnum(f);

        /* set new value for @wait queue to wake up on */
        if(queue == Q_WAIT) {
          new->wait=getlong(f);
          if(!next_wait || new->wait < next_wait)
            next_wait=new->wait;
        }

        for(i=0;i<10;i++) {
          if(*(s=getstring_noalloc(f)))
            SPTR(new->env[i], s);
          else
            new->env[i]=NULL;
        }

        if(last)
          last->next=new;
        else
          (queue == Q_COMMAND)?qfirst:qwait = new;
        last=new;
        count++;
      }

      /* set the tail pointer */
      (queue == Q_COMMAND)?qlast:qwaitlast = last;
      break;

    }

    queue_size[queue]=count;
  }
}

/* restores all players' queue quota counts from a @reboot */
void restore_queue_quotas()
{
  struct bque *ptr;
  int queue;

  for(queue=0;queue<2;queue++)
    for(ptr=(queue == Q_COMMAND)?qfirst:qwait; ptr; ptr=ptr->next)
      db[ptr->player].queue++;
}
