/* hdrs/db.h */

extern dbref speaker;
extern dbref notify_location;

extern dbref object_atime;
extern int dberrcheck;
extern int boot_flags;
extern int mare_test_db;
extern int db_forceload;
extern int db_readonly;

extern int default_color;

extern int ndisrooms;
extern int treedepth;

extern int enter_plane;

extern FILE *db_read_file;
extern char *db_file;
extern int dbref_len;
extern int db_attrs;
extern int db_lists;

extern int disp_hidden;

extern char *wptr[10];
extern char log_text[8192];
extern dbref log_dbref;


/* Object type definitions (0-7) */

#define TYPE_ROOM 	0
#define TYPE_THING 	1
#define TYPE_EXIT 	2
#define TYPE_PLAYER	3
#define TYPE_ZONE	4
#define TYPE_GARBAGE	7

#define NOTYPE		-1	// no particular type

#define Typeof(x) (db[x].flags & 0x7)
#define IS(thing,type,flag) ((Typeof(thing)==type) && (db[thing].flags&(flag)))

/* Miscellaneous Flags (Any Object Type) */
#define CHOWN_OK	BV5
#define DARK		BV6
#define VERBOSE		BV8
#define HAVEN		BV10
#define GOING		BV14
#define PUPPET		BV17
#define VISIBLE		BV20
#define OPAQUE		BV23
#define QUIET		BV24
#define LIGHT		BV26
#define BEARING		BV27

/* The above flags OR'd together */
#define NOTYPE_FLAGMASK	\
  (BV5|BV6|BV8|BV10|BV14|BV17|BV20|BV23|BV24|BV26|BV27)

/* Object flags */
#define THING_MOUNT        BV3
#define THING_UNSAVED      BV7
#define THING_DESTROY_OK   BV9
#define THING_WANDER       BV11
#define THING_ENTER_OK     BV15
#define THING_MONSTER      BV22
#define THING_GRAB_OK      BV25

/* Exit flags */
#define EXIT_TRANSPARENT   BV3
#define EXIT_DOOR          BV4

/* Player flags */
#define PLAYER_BUILD       BV4
#define PLAYER_SLAVE       BV7
#define PLAYER_CONNECT     BV9
#define PLAYER_INACTIVE_OK BV11
#define PLAYER_TERSE       BV22
#define PLAYER_NOTICE      BV25

/* Room flags */
#define ROOM_XZONE	BV3
#define ROOM_OCEANIC	BV4
#define ROOM_WALL	BV7
#define ROOM_JUMP_OK	BV9
#define ROOM_HEALING	BV11
#define ROOM_FLOATING	BV12
#define ROOM_ILLNESS	BV13
#define ROOM_NEST	BV15
#define ROOM_FOG	BV16
#define ROOM_LINK_OK	BV18
#define ROOM_ARENA	BV19
#define ROOM_SANCTUARY	BV21
#define ROOM_SHAFT	BV22
#define ROOM_ABODE	BV25

/* Zone flags */
#define ZONE_UNIVERSAL	BV4
#define ZONE_RESTRICTED	BV7

/* Internal (unsaved) flags */
#define ROOM_MARKED	BV31	// Check for disconnected rooms only


/* special dbref's */
#define GOD		 1	// ultimate wizard character
#define NOTHING		-1	// null dbref
#define AMBIGUOUS	-2	// multiple possibilities, for matchers
#define HOME		-3	// virtual room, represents mover's home

/* structure for defining flags in game/unparse.c */
extern struct flagstr {
  char *name;
  char flag;
  char objtype;
  int bits;
} flaglist[];


/* macros for gender substitutions */
enum gender_types { MALE, FEMALE, NEUTER, PLURAL };
enum pronoun_types { SUBJ=0, OBJN=4, POSS=8, APOSS=12 };

/* returns a random number from 0 to num-1 */
#define rand_num(num) ({ int _n=num; \
                         (_n < 2)?0:((mt_rand() & 0x7fffffff) % _n); })

/* randomized division with floating point effect */
#define rdiv(x,y)     ({ typeof(x) _x=x; int _y=y; \
                         (!_y)?0:((_x / _y)+(rand_num(_y) < _x % _y)); })

/* get size of file */
#define fsize(file)   ({ struct stat st; \
                         stat(file, &st)?-1:st.st_size; })

/* modification time of file */
#define fdate(file)   ({ struct stat st; \
                         stat(file, &st)?-1:st.st_mtime; })


/* read/write strings/integers to a file in big-endian format: */

/* strings */
#define getstring(x,p) SPTR(p, getstring_noalloc(x))

/* 8-bit */
#define putchr(f, num) fputc(num, f)

/* 16-bit */
#define putshort(f, num)  ({ unsigned short _x=htons(num); \
			     fwrite(&_x, 2, 1, f); })

#define getshort(f)       ({ unsigned short _x; \
                             fread(&_x, 2, 1, f); ntohs(_x); })

/* 32-bit */
#define putnum(f, num)  ({ int _x=htonl(num); fwrite(&_x, 4, 1, f); })

#define getnum(f)       ({ int _x; fread(&_x, 4, 1, f); ntohl(_x); })


/* 64-bit */
#define putlong(f, num)  ({ quad _x=htonll(num); fwrite(&_x, 8, 1, f); })

#define getlong(f)       ({ quad _x; fread(&_x, 8, 1, f); ntohll(_x); })


#define atr_get(thing, num) atr_get_obj(thing, NOTHING, num)
#define atr_add(thing, num, str) atr_add_obj(thing, NOTHING, num, str)

#define class(x) (db[x].data->class)
#define terminfo(x,y) (db[db[x].owner].data->term & (y))

#define power(x,pow) controls(x, NOTHING, pow)
#define could_doit(x,y,attr) ({ dbref _tr=y; char *_s=atr_get(_tr, attr); \
                                *_s?eval_boolexp(x, _tr, _s):1; })
#define notify_except(loc,thing,fmt,args...) \
                         notify_room(loc, thing, NOTHING, fmt, ## args)
#define atr_parse(x,y,attr) parse_function(x, y, atr_get(x, attr))
#define get_zone(player) ({ dbref *_zone=db[mainroom(player)].zone; \
                            _zone ? *_zone : NOTHING; })

#define Invalid(x) ({ dbref _obj=x; (_obj < 0 || _obj >= db_top); })
#define Immortal(x) (db[db[x].owner].data->class >= CLASS_IMMORTAL)
#define Guest(x) (db[db[x].owner].data->class == CLASS_GUEST)
#define Builder(x) (IS(db[x].owner, TYPE_PLAYER, PLAYER_BUILD) || \
                    power(x, POW_BUILD))
#define Quiet(x) ((db[x].flags & QUIET) || (Typeof(x) != TYPE_PLAYER && \
                  !(db[x].flags & PUPPET)))

#define Desc(x) (db[x].data->desc)
#define ipaddr(x) ({ struct in_addr _addr; _addr.s_addr=x; inet_ntoa(_addr); })

/* game startup flags */
#define BOOT_RESTART	BV0	// Hard reset the game during a reboot
#define BOOT_QUIET	BV1	// Suppress log_main from being displayed
#define BOOT_VALIDATE	BV2	// Do not save semaphores in the database

/* database format versions */
#define F_MUSH		1	// TinyMush
#define F_MUSE		2	// TinyMuse
#define F_MUSE_ZONED	3	// TinyMuse Zoned
#define F_MARE		4	// TinyMare Compressed RAM
#define F_MUSH3		5	// TinyMush-3

/* tinymare db flags */
#define DB_LONGINT	BV0	// Time values stored as 8 bytes instead of 4
#define DB_COMBAT	BV1	// Combat data is stored for players

/* tinymush db flags */
#define MUSH_ZONE	BV0	// ZONE/DOMAIN field
#define MUSH_LINK	BV1	// LINK field (exits from objs)
#define MUSH_GDBM	BV2	// attrs are in a gdbm db, not here
#define MUSH_ATRNAME	BV3	// NAME is an attr, not in the hdr
#define MUSH_ATRKEY	BV4	// KEY is an attr, not in the hdr
#define MUSH_PARENT	BV5	// db has the PARENT field
#define MUSH_ATRMONEY	BV7	// Money is kept in an attribute
#define MUSH_XFLAGS	BV8	// An extra word of flags
#define MUSH_POWERS	BV9	// Powers?
#define MUSH_3FLAGS	BV10	// Adding a 3rd flag word
#define MUSH_QUOTED	BV11	// Quoted strings, ala PennMUSH

/* attribute structure */
typedef struct {
  char *name;		  // full name of attribute
  dbref obj;		  // object in which attr is defined
  unsigned char num;	  // attr number in cache (1-255)
  unsigned int flags:16;  // any flags the attribute holds (listed below)
} ATTR;

/* attribute flags */
#define AF_OSEE    BV0    // players other than owner can see it
#define AF_DARK    BV1    // wizards or owner of object defined on can see
#define AF_WIZ     BV2    // only wizards can change it
#define AF_UNIMP   BV3    // not important -- don't save value in the db
#define AF_HIDDEN  BV4    // changed and viewed only by the high almighty
#define AF_DATE    BV5    // date stored in universal longint form
#define INH        BV6    // value inherited by children
#define AF_LOCK    BV7    // interpreted as a boolean expression
#define AF_FUNC    BV8    // this is a user defined function
#define AF_HAVEN   BV9    // can't execute commands on this attribute
#define AF_BITMAP  BV10   // an attribute expressed in hexadecimal form
#define AF_DBREF   BV11   // calls unparse_object on #dbref numbers
#define AF_COMBAT  BV12   // attribute editable only with pow_combat

extern const char *attr_options[16];

/* external attribute list */
enum attributes {
#define DOATTR(num, var, name, flags) var=num,
#include "attrib.h"
#undef DOATTR
};


/* attribute list structure */
struct alist {
  struct alist *next;
  dbref obj;
  unsigned char num;
  char text[0];
};

struct atrdef {
  struct atrdef *next;
  ATTR atr;
};

typedef struct atrdef ATRDEF;
typedef struct alist ALIST;


/* database structures */

struct playerdata {
  DESC *desc;

  unsigned int class:8;
  unsigned int rows:8;
  unsigned int cols:8;

  int tz:5;
  unsigned int tzdst:1;
  unsigned int gender:1;
  unsigned int passtype:1;

  unsigned int term:16;

  int steps;
  int sessions;
  int age;

  long last;
  long lastoff;

  unsigned char pass[20];
  unsigned char powers[(NUM_POWS+3)/4];
};
  
/* database reference object data */
struct object {
  char *name;			// Name of object
  dbref location;		// Pointer to container
  dbref contents;		// Pointer to first item
  dbref exits;			// Pointer to first exit for rooms
  dbref link;			// Pointer to home/destination
  dbref next;			// Pointer to next in contents/exits chain
  dbref owner;			// Who controls this object

  unsigned int flags;		// Object flag list
  int plane;			// Dimension plane setting
  int pennies;			// Money value

  int queue;			// Number of commands in queue on object

  ALIST *attrs;			// Attribute data list
  ATRDEF *atrdefs;		// User-defined attribute list

  dbref *parents;		// Parents list
  dbref *children;		// Children list
  dbref *zone;			// Zone list

  struct playerdata *data;	// Player data (see structure above)

  long mod_time;		// Date of modification
  long create_time;		// Date of creation
};

extern struct object *db;
extern dbref db_top;
extern dbref *uzo;

/* macros for object lists */

#define DOLIST(var, first) \
  for(var=first; var != NOTHING; var=db[var].next)
#define ADDLIST(thing, locative) \
  (db[thing].next=locative, locative=thing)

#define PUSH_F(list,value)	push_first(&list,value)
#define PUSH_L(list,value)	push_last(&list,value)
#define PULL(list,value)	pull_list(&list,value)

/* returns true if <match> is in <list> */
#define inlist(list,match) \
  ({ \
    dbref *_l=list; \
    while(_l && *_l != NOTHING && *_l != (match)) _l++; \
    (!_l || *_l == NOTHING)?NULL:_l; \
  })

/* erases contents of <list> */
#define CLEAR(list)  ({ dbref **_l=&(list); \
                        if(*_l) { free(*_l); *_l=NULL; db_lists--; } })

/* attribute list structure */
struct all_atr_list {
  ATTR *type;
  char *value;
  unsigned int numinherit;
};

/* TinyMare-specific db configuration rules */
extern struct db_rule {
  int type;
  char *text;
  int (*load)();
  void (*save)();
} db_rule[];

/* hash table manipulation */
typedef struct hashent {
  struct hashent *ptr;
} hash;
